@@ -34,6 +34,7 @@
 #![allow(clippy::reversed_empty_ranges)]
 
 use crate::{
+    binding_model::buffer_binding_type_alignment,
     command::{
         BasePass, DrawError, MapPassErr, PassErrorScope, RenderCommand, RenderCommandError,
         StateChange,
@@ -46,7 +47,7 @@ use crate::{
     error::{ErrorFormatter, PrettyError},
     hub::{GlobalIdentityHandlerFactory, HalApi, Hub, Resource, Storage, Token},
     id,
-    memory_init_tracker::{MemoryInitKind, MemoryInitTrackerAction},
+    init_tracker::{BufferInitTrackerAction, MemoryInitKind},
     pipeline::PipelineFlags,
     track::{TrackerSet, UsageConflict},
     validation::check_buffer_usage,
@@ -96,7 +97,11 @@ impl RenderBundleEncoder {
             parent_id,
             context: RenderPassContext {
                 attachments: AttachmentData {
-                    colors: desc.color_formats.iter().cloned().collect(),
+                    colors: if desc.color_formats.len() > hal::MAX_COLOR_TARGETS {
+                        return Err(CreateRenderBundleError::TooManyColorAttachments);
+                    } else {
+                        desc.color_formats.iter().cloned().collect()
+                    },
                     resolves: ArrayVec::new(),
                     depth_stencil: desc.depth_stencil.map(|ds| ds.format),
                 },
@@ -196,15 +201,6 @@ impl RenderBundleEncoder {
 
                     let offsets = &base.dynamic_offsets[..num_dynamic_offsets as usize];
                     base.dynamic_offsets = &base.dynamic_offsets[num_dynamic_offsets as usize..];
-                    // Check for misaligned offsets.
-                    if let Some(offset) = offsets
-                        .iter()
-                        .map(|offset| *offset as wgt::BufferAddress)
-                        .find(|offset| offset % wgt::BIND_BUFFER_ALIGNMENT != 0)
-                    {
-                        return Err(RenderCommandError::UnalignedBufferOffset(offset))
-                            .map_pass_err(scope);
-                    }
 
                     let bind_group = state
                         .trackers
@@ -220,6 +216,22 @@ impl RenderBundleEncoder {
                         .map_pass_err(scope);
                     }
 
+                    // Check for misaligned offsets.
+                    for (offset, info) in offsets
+                        .iter()
+                        .map(|offset| *offset as wgt::BufferAddress)
+                        .zip(bind_group.dynamic_binding_info.iter())
+                    {
+                        let (alignment, limit_name) =
+                            buffer_binding_type_alignment(&device.limits, info.binding_type);
+                        if offset % alignment as u64 != 0 {
+                            return Err(RenderCommandError::UnalignedBufferOffset(
+                                offset, limit_name, alignment,
+                            ))
+                            .map_pass_err(scope);
+                        }
+                    }
+
                     buffer_memory_init_actions.extend_from_slice(&bind_group.used_buffer_ranges);
 
                     state.set_bind_group(index, bind_group_id, bind_group.layout_id, offsets);
@@ -287,11 +299,11 @@ impl RenderBundleEncoder {
                         Some(s) => offset + s.get(),
                         None => buffer.size,
                     };
-                    buffer_memory_init_actions.push(MemoryInitTrackerAction {
-                        id: buffer_id,
-                        range: offset..end,
-                        kind: MemoryInitKind::NeedsInitializedMemory,
-                    });
+                    buffer_memory_init_actions.extend(buffer.initialization_status.create_action(
+                        buffer_id,
+                        offset..end,
+                        MemoryInitKind::NeedsInitializedMemory,
+                    ));
                     state.index.set_format(index_format);
                     state.index.set_buffer(buffer_id, offset..end);
                 }
@@ -314,11 +326,11 @@ impl RenderBundleEncoder {
                         Some(s) => offset + s.get(),
                         None => buffer.size,
                     };
-                    buffer_memory_init_actions.push(MemoryInitTrackerAction {
-                        id: buffer_id,
-                        range: offset..end,
-                        kind: MemoryInitKind::NeedsInitializedMemory,
-                    });
+                    buffer_memory_init_actions.extend(buffer.initialization_status.create_action(
+                        buffer_id,
+                        offset..end,
+                        MemoryInitKind::NeedsInitializedMemory,
+                    ));
                     state.vertex[slot as usize].set_buffer(buffer_id, offset..end);
                 }
                 RenderCommand::SetPushConstant {
@@ -435,18 +447,11 @@ impl RenderBundleEncoder {
                     check_buffer_usage(buffer.usage, wgt::BufferUsages::INDIRECT)
                         .map_pass_err(scope)?;
 
-                    buffer_memory_init_actions.extend(
-                        buffer
-                            .initialization_status
-                            .check(
-                                offset..(offset + mem::size_of::<wgt::DrawIndirectArgs>() as u64),
-                            )
-                            .map(|range| MemoryInitTrackerAction {
-                                id: buffer_id,
-                                range,
-                                kind: MemoryInitKind::NeedsInitializedMemory,
-                            }),
-                    );
+                    buffer_memory_init_actions.extend(buffer.initialization_status.create_action(
+                        buffer_id,
+                        offset..(offset + mem::size_of::<wgt::DrawIndirectArgs>() as u64),
+                        MemoryInitKind::NeedsInitializedMemory,
+                    ));
 
                     commands.extend(state.flush_vertices());
                     commands.extend(state.flush_binds());
@@ -476,18 +481,11 @@ impl RenderBundleEncoder {
                     check_buffer_usage(buffer.usage, wgt::BufferUsages::INDIRECT)
                         .map_pass_err(scope)?;
 
-                    buffer_memory_init_actions.extend(
-                        buffer
-                            .initialization_status
-                            .check(
-                                offset..(offset + mem::size_of::<wgt::DrawIndirectArgs>() as u64),
-                            )
-                            .map(|range| MemoryInitTrackerAction {
-                                id: buffer_id,
-                                range,
-                                kind: MemoryInitKind::NeedsInitializedMemory,
-                            }),
-                    );
+                    buffer_memory_init_actions.extend(buffer.initialization_status.create_action(
+                        buffer_id,
+                        offset..(offset + mem::size_of::<wgt::DrawIndirectArgs>() as u64),
+                        MemoryInitKind::NeedsInitializedMemory,
+                    ));
 
                     commands.extend(state.index.flush());
                     commands.extend(state.flush_vertices());
@@ -551,6 +549,8 @@ impl RenderBundleEncoder {
 pub enum CreateRenderBundleError {
     #[error("invalid number of samples {0}")]
     InvalidSampleCount(u32),
+    #[error("number of color attachments exceeds the limit")]
+    TooManyColorAttachments,
 }
 
 /// Error type returned from `RenderBundleEncoder::new` if the sample count is invalid.
@@ -586,7 +586,7 @@ pub struct RenderBundle {
     pub(super) is_ds_read_only: bool,
     pub(crate) device_id: Stored<id::DeviceId>,
     pub(crate) used: TrackerSet,
-    pub(super) buffer_memory_init_actions: Vec<MemoryInitTrackerAction<id::BufferId>>,
+    pub(super) buffer_memory_init_actions: Vec<BufferInitTrackerAction>,
     pub(super) context: RenderPassContext,
     pub(crate) life_guard: LifeGuard,
 }
