@@ -6,7 +6,7 @@ use crate::{
     error::{ErrorFormatter, PrettyError},
     hub::{Global, GlobalIdentityHandlerFactory, HalApi, Storage, Token},
     id::{BufferId, CommandEncoderId, TextureId},
-    memory_init_tracker::{MemoryInitKind, MemoryInitTrackerAction},
+    init_tracker::MemoryInitKind,
     resource::{Texture, TextureErrorDimension},
     track::TextureSelector,
 };
@@ -453,26 +453,20 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         }
 
         // Make sure source is initialized memory and mark dest as initialized.
-        cmd_buf.buffer_memory_init_actions.extend(
-            dst_buffer
-                .initialization_status
-                .check(destination_offset..(destination_offset + size))
-                .map(|range| MemoryInitTrackerAction {
-                    id: destination,
-                    range,
-                    kind: MemoryInitKind::ImplicitlyInitialized,
-                }),
-        );
-        cmd_buf.buffer_memory_init_actions.extend(
-            src_buffer
-                .initialization_status
-                .check(source_offset..(source_offset + size))
-                .map(|range| MemoryInitTrackerAction {
-                    id: source,
-                    range,
-                    kind: MemoryInitKind::NeedsInitializedMemory,
-                }),
-        );
+        cmd_buf
+            .buffer_memory_init_actions
+            .extend(dst_buffer.initialization_status.create_action(
+                destination,
+                destination_offset..(destination_offset + size),
+                MemoryInitKind::ImplicitlyInitialized,
+            ));
+        cmd_buf
+            .buffer_memory_init_actions
+            .extend(src_buffer.initialization_status.create_action(
+                source,
+                source_offset..(source_offset + size),
+                MemoryInitKind::NeedsInitializedMemory,
+            ));
 
         let region = hal::BufferCopy {
             src_offset: source_offset,
@@ -573,16 +567,13 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             true,
         )?;
 
-        cmd_buf.buffer_memory_init_actions.extend(
-            src_buffer
-                .initialization_status
-                .check(source.layout.offset..(source.layout.offset + required_buffer_bytes_in_copy))
-                .map(|range| MemoryInitTrackerAction {
-                    id: source.buffer,
-                    range,
-                    kind: MemoryInitKind::NeedsInitializedMemory,
-                }),
-        );
+        cmd_buf
+            .buffer_memory_init_actions
+            .extend(src_buffer.initialization_status.create_action(
+                source.buffer,
+                source.layout.offset..(source.layout.offset + required_buffer_bytes_in_copy),
+                MemoryInitKind::NeedsInitializedMemory,
+            ));
 
         if !conv::is_valid_copy_dst_texture_format(dst_texture.desc.format) {
             return Err(
@@ -703,19 +694,14 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             );
         }
 
-        cmd_buf.buffer_memory_init_actions.extend(
-            dst_buffer
-                .initialization_status
-                .check(
-                    destination.layout.offset
-                        ..(destination.layout.offset + required_buffer_bytes_in_copy),
-                )
-                .map(|range| MemoryInitTrackerAction {
-                    id: destination.buffer,
-                    range,
-                    kind: MemoryInitKind::ImplicitlyInitialized,
-                }),
-        );
+        cmd_buf
+            .buffer_memory_init_actions
+            .extend(dst_buffer.initialization_status.create_action(
+                destination.buffer,
+                destination.layout.offset
+                    ..(destination.layout.offset + required_buffer_bytes_in_copy),
+                MemoryInitKind::ImplicitlyInitialized,
+            ));
 
         let regions = (0..array_layer_count).map(|rel_array_layer| {
             let mut texture_base = src_base.clone();
