@@ -17,22 +17,22 @@ use hal::Device as _;
 use parking_lot::Mutex;
 use thiserror::Error;
 
-use std::{mem, sync::atomic::Ordering};
+use std::mem;
 
 /// A struct that keeps lists of resources that are no longer needed by the user.
 #[derive(Debug, Default)]
 pub(super) struct SuspectedResources {
-    pub(crate) buffers: Vec<id::Valid<id::BufferId>>,
-    pub(crate) textures: Vec<id::Valid<id::TextureId>>,
-    pub(crate) texture_views: Vec<id::Valid<id::TextureViewId>>,
-    pub(crate) samplers: Vec<id::Valid<id::SamplerId>>,
-    pub(crate) bind_groups: Vec<id::Valid<id::BindGroupId>>,
-    pub(crate) compute_pipelines: Vec<id::Valid<id::ComputePipelineId>>,
-    pub(crate) render_pipelines: Vec<id::Valid<id::RenderPipelineId>>,
-    pub(crate) bind_group_layouts: Vec<id::Valid<id::BindGroupLayoutId>>,
-    pub(crate) pipeline_layouts: Vec<Stored<id::PipelineLayoutId>>,
-    pub(crate) render_bundles: Vec<id::Valid<id::RenderBundleId>>,
-    pub(crate) query_sets: Vec<id::Valid<id::QuerySetId>>,
+    pub(super) buffers: Vec<id::Valid<id::BufferId>>,
+    pub(super) textures: Vec<id::Valid<id::TextureId>>,
+    pub(super) texture_views: Vec<id::Valid<id::TextureViewId>>,
+    pub(super) samplers: Vec<id::Valid<id::SamplerId>>,
+    pub(super) bind_groups: Vec<id::Valid<id::BindGroupId>>,
+    pub(super) compute_pipelines: Vec<id::Valid<id::ComputePipelineId>>,
+    pub(super) render_pipelines: Vec<id::Valid<id::RenderPipelineId>>,
+    pub(super) bind_group_layouts: Vec<id::Valid<id::BindGroupLayoutId>>,
+    pub(super) pipeline_layouts: Vec<Stored<id::PipelineLayoutId>>,
+    pub(super) render_bundles: Vec<id::Valid<id::RenderBundleId>>,
+    pub(super) query_sets: Vec<id::Valid<id::QuerySetId>>,
 }
 
 impl SuspectedResources {
@@ -128,35 +128,65 @@ impl<A: hal::Api> NonReferencedResources<A> {
     }
 
     unsafe fn clean(&mut self, device: &A::Device) {
-        for raw in self.buffers.drain(..) {
-            device.destroy_buffer(raw);
+        if !self.buffers.is_empty() {
+            profiling::scope!("destroy_buffers");
+            for raw in self.buffers.drain(..) {
+                device.destroy_buffer(raw);
+            }
         }
-        for raw in self.textures.drain(..) {
-            device.destroy_texture(raw);
+        if !self.textures.is_empty() {
+            profiling::scope!("destroy_textures");
+            for raw in self.textures.drain(..) {
+                device.destroy_texture(raw);
+            }
         }
-        for (_, raw) in self.texture_views.drain(..) {
-            device.destroy_texture_view(raw);
+        if !self.texture_views.is_empty() {
+            profiling::scope!("destroy_texture_views");
+            for (_, raw) in self.texture_views.drain(..) {
+                device.destroy_texture_view(raw);
+            }
         }
-        for raw in self.samplers.drain(..) {
-            device.destroy_sampler(raw);
+        if !self.samplers.is_empty() {
+            profiling::scope!("destroy_samplers");
+            for raw in self.samplers.drain(..) {
+                device.destroy_sampler(raw);
+            }
         }
-        for raw in self.bind_groups.drain(..) {
-            device.destroy_bind_group(raw);
+        if !self.bind_groups.is_empty() {
+            profiling::scope!("destroy_bind_groups");
+            for raw in self.bind_groups.drain(..) {
+                device.destroy_bind_group(raw);
+            }
         }
-        for raw in self.compute_pipes.drain(..) {
-            device.destroy_compute_pipeline(raw);
+        if !self.compute_pipes.is_empty() {
+            profiling::scope!("destroy_compute_pipelines");
+            for raw in self.compute_pipes.drain(..) {
+                device.destroy_compute_pipeline(raw);
+            }
         }
-        for raw in self.render_pipes.drain(..) {
-            device.destroy_render_pipeline(raw);
+        if !self.render_pipes.is_empty() {
+            profiling::scope!("destroy_render_pipelines");
+            for raw in self.render_pipes.drain(..) {
+                device.destroy_render_pipeline(raw);
+            }
         }
-        for raw in self.bind_group_layouts.drain(..) {
-            device.destroy_bind_group_layout(raw);
+        if !self.bind_group_layouts.is_empty() {
+            profiling::scope!("destroy_bind_group_layouts");
+            for raw in self.bind_group_layouts.drain(..) {
+                device.destroy_bind_group_layout(raw);
+            }
         }
-        for raw in self.pipeline_layouts.drain(..) {
-            device.destroy_pipeline_layout(raw);
+        if !self.pipeline_layouts.is_empty() {
+            profiling::scope!("destroy_pipeline_layouts");
+            for raw in self.pipeline_layouts.drain(..) {
+                device.destroy_pipeline_layout(raw);
+            }
         }
-        for raw in self.query_sets.drain(..) {
-            device.destroy_query_set(raw);
+        if !self.query_sets.is_empty() {
+            profiling::scope!("destroy_query_sets");
+            for raw in self.query_sets.drain(..) {
+                device.destroy_query_set(raw);
+            }
         }
     }
 }
@@ -290,7 +320,7 @@ impl<A: hal::Api> LifetimeTracker<A> {
     }
 
     pub fn cleanup(&mut self, device: &A::Device) {
-        profiling::scope!("cleanup");
+        profiling::scope!("cleanup", "LifetimeTracker");
         unsafe {
             self.free_resources.clean(device);
         }
@@ -367,7 +397,7 @@ impl<A: HalApi> LifetimeTracker<A> {
                     if let Some(res) = hub.bind_groups.unregister_locked(id.0, &mut *guard) {
                         self.suspected_resources.add_trackers(&res.used);
 
-                        let submit_index = res.life_guard.submission_index.load(Ordering::Acquire);
+                        let submit_index = res.life_guard.life_count();
                         self.active
                             .iter_mut()
                             .find(|a| a.index == submit_index)
@@ -393,7 +423,7 @@ impl<A: HalApi> LifetimeTracker<A> {
 
                     if let Some(res) = hub.texture_views.unregister_locked(id.0, &mut *guard) {
                         self.suspected_resources.textures.push(res.parent_id.value);
-                        let submit_index = res.life_guard.submission_index.load(Ordering::Acquire);
+                        let submit_index = res.life_guard.life_count();
                         self.active
                             .iter_mut()
                             .find(|a| a.index == submit_index)
@@ -418,7 +448,7 @@ impl<A: HalApi> LifetimeTracker<A> {
                     }
 
                     if let Some(res) = hub.textures.unregister_locked(id.0, &mut *guard) {
-                        let submit_index = res.life_guard.submission_index.load(Ordering::Acquire);
+                        let submit_index = res.life_guard.life_count();
                         let raw = match res.inner {
                             resource::TextureInner::Native { raw: Some(raw) } => raw,
                             _ => continue,
@@ -446,7 +476,7 @@ impl<A: HalApi> LifetimeTracker<A> {
                     }
 
                     if let Some(res) = hub.samplers.unregister_locked(id.0, &mut *guard) {
-                        let submit_index = res.life_guard.submission_index.load(Ordering::Acquire);
+                        let submit_index = res.life_guard.life_count();
                         self.active
                             .iter_mut()
                             .find(|a| a.index == submit_index)
@@ -471,7 +501,7 @@ impl<A: HalApi> LifetimeTracker<A> {
                     log::debug!("Buffer {:?} is detached", id);
 
                     if let Some(res) = hub.buffers.unregister_locked(id.0, &mut *guard) {
-                        let submit_index = res.life_guard.submission_index.load(Ordering::Acquire);
+                        let submit_index = res.life_guard.life_count();
                         if let resource::BufferMapState::Init { stage_buffer, .. } = res.map_state {
                             self.free_resources.buffers.push(stage_buffer);
                         }
@@ -498,7 +528,7 @@ impl<A: HalApi> LifetimeTracker<A> {
                     }
 
                     if let Some(res) = hub.compute_pipelines.unregister_locked(id.0, &mut *guard) {
-                        let submit_index = res.life_guard.submission_index.load(Ordering::Acquire);
+                        let submit_index = res.life_guard.life_count();
                         self.active
                             .iter_mut()
                             .find(|a| a.index == submit_index)
@@ -522,7 +552,7 @@ impl<A: HalApi> LifetimeTracker<A> {
                     }
 
                     if let Some(res) = hub.render_pipelines.unregister_locked(id.0, &mut *guard) {
-                        let submit_index = res.life_guard.submission_index.load(Ordering::Acquire);
+                        let submit_index = res.life_guard.life_count();
                         self.active
                             .iter_mut()
                             .find(|a| a.index == submit_index)
@@ -588,7 +618,7 @@ impl<A: HalApi> LifetimeTracker<A> {
                     // #[cfg(feature = "trace")]
                     // trace.map(|t| t.lock().add(trace::Action::DestroyComputePipeline(id.0)));
                     if let Some(res) = hub.query_sets.unregister_locked(id.0, &mut *guard) {
-                        let submit_index = res.life_guard.submission_index.load(Ordering::Acquire);
+                        let submit_index = res.life_guard.life_count();
                         self.active
                             .iter_mut()
                             .find(|a| a.index == submit_index)
@@ -615,7 +645,7 @@ impl<A: HalApi> LifetimeTracker<A> {
             let resource_id = stored.value;
             let buf = &buffer_guard[resource_id];
 
-            let submit_index = buf.life_guard.submission_index.load(Ordering::Acquire);
+            let submit_index = buf.life_guard.life_count();
             log::trace!(
                 "Mapping of {:?} at submission {:?} gets assigned to active {:?}",
                 resource_id,
