@@ -59,6 +59,7 @@ pub enum Action<'a> {
         parent_id: id::SurfaceId,
     },
     Present(id::SurfaceId),
+    DiscardSurfaceTexture(id::SurfaceId),
     CreateBindGroupLayout(
         id::BindGroupLayoutId,
         crate::binding_model::BindGroupLayoutDescriptor<'a>,
@@ -151,7 +152,7 @@ pub enum Command {
         offset: wgt::BufferAddress,
         size: Option<wgt::BufferSize>,
     },
-    ClearImage {
+    ClearTexture {
         dst: id::TextureId,
         subresource_range: wgt::ImageSubresourceRange,
     },
