@@ -17,21 +17,12 @@ use std::cell::Cell;
 use std::{fmt::Debug, marker::PhantomData, mem, ops};
 
 /// A simple structure to manage identities of objects.
-#[derive(Debug)]
+#[derive(Debug, Default)]
 pub struct IdentityManager {
     free: Vec<Index>,
     epochs: Vec<Epoch>,
 }
 
-impl Default for IdentityManager {
-    fn default() -> Self {
-        Self {
-            free: Default::default(),
-            epochs: Default::default(),
-        }
-    }
-}
-
 impl IdentityManager {
     pub fn from_index(min_index: u32) -> Self {
         Self {
@@ -110,7 +101,7 @@ impl<T, I: id::TypedId> ops::IndexMut<id::Valid<I>> for Storage<T, I> {
 }
 
 impl<T, I: id::TypedId> Storage<T, I> {
-    pub(crate) fn _contains(&self, id: I) -> bool {
+    pub(crate) fn contains(&self, id: I) -> bool {
         let (index, epoch, _) = id.unzip();
         match self.map[index as usize] {
             Element::Vacant => false,
@@ -644,14 +635,12 @@ impl<A: HalApi, F: GlobalIdentityHandlerFactory> Hub<A, F> {
                 }
             }
         }
-        {
-            let textures = self.textures.data.read();
-            for element in self.texture_views.data.write().map.drain(..) {
-                if let Element::Occupied(texture_view, _) = element {
-                    let device = &devices[textures[texture_view.parent_id.value].device_id.value];
-                    unsafe {
-                        device.raw.destroy_texture_view(texture_view.raw);
-                    }
+
+        for element in self.texture_views.data.write().map.drain(..) {
+            if let Element::Occupied(texture_view, _) = element {
+                let device = &devices[texture_view.device_id.value];
+                unsafe {
+                    device.raw.destroy_texture_view(texture_view.raw);
                 }
             }
         }
@@ -1033,6 +1022,7 @@ impl HalApi for hal::api::Dx11 {
 impl HalApi for hal::api::Gles {
     const VARIANT: Backend = Backend::Gl;
     fn create_instance_from_hal(name: &str, hal_instance: Self::Instance) -> Instance {
+        #[allow(clippy::needless_update)]
         Instance {
             name: name.to_owned(),
             gl: Some(hal_instance),
