@@ -1,18 +1,16 @@
-/* This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-
 mod buffer;
 mod range;
 mod texture;
 
 use crate::{
-    conv, hub,
+    hub,
     id::{self, TypedId, Valid},
     resource, Epoch, FastHashMap, Index, RefCount,
 };
 
-use std::{collections::hash_map::Entry, fmt, marker::PhantomData, ops, vec::Drain};
+use std::{
+    collections::hash_map::Entry, fmt, marker::PhantomData, num::NonZeroU32, ops, vec::Drain,
+};
 use thiserror::Error;
 
 pub(crate) use buffer::BufferState;
@@ -78,14 +76,6 @@ pub(crate) trait ResourceState: Clone + Default {
         output: Option<&mut Vec<PendingTransition<Self>>>,
     ) -> Result<(), PendingTransition<Self>>;
 
-    /// Sets up the first usage of the selected sub-resources.
-    fn prepend(
-        &mut self,
-        id: Valid<Self::Id>,
-        selector: Self::Selector,
-        usage: Self::Usage,
-    ) -> Result<(), PendingTransition<Self>>;
-
     /// Merge the state of this resource tracked by a different instance
     /// with the current one.
     ///
@@ -125,19 +115,16 @@ pub(crate) struct PendingTransition<S: ResourceState> {
 }
 
 impl PendingTransition<BufferState> {
-    /// Produce the gfx-hal barrier corresponding to the transition.
-    pub fn into_hal<'a, B: hal::Backend>(
+    /// Produce the hal barrier corresponding to the transition.
+    pub fn into_hal<'a, A: hal::Api>(
         self,
-        buf: &'a resource::Buffer<B>,
-    ) -> hal::memory::Barrier<'a, B> {
+        buf: &'a resource::Buffer<A>,
+    ) -> hal::BufferBarrier<'a, A> {
         log::trace!("\tbuffer -> {:?}", self);
-        let &(ref target, _) = buf.raw.as_ref().expect("Buffer is destroyed");
-        hal::memory::Barrier::Buffer {
-            states: conv::map_buffer_state(self.usage.start)
-                ..conv::map_buffer_state(self.usage.end),
-            target,
-            range: hal::buffer::SubRange::WHOLE,
-            families: None,
+        let buffer = buf.raw.as_ref().expect("Buffer is destroyed");
+        hal::BufferBarrier {
+            buffer,
+            usage: self.usage,
         }
     }
 }
@@ -152,26 +139,27 @@ impl From<PendingTransition<BufferState>> for UsageConflict {
 }
 
 impl PendingTransition<TextureState> {
-    /// Produce the gfx-hal barrier corresponding to the transition.
-    pub fn into_hal<'a, B: hal::Backend>(
+    /// Produce the hal barrier corresponding to the transition.
+    pub fn into_hal<'a, A: hal::Api>(
         self,
-        tex: &'a resource::Texture<B>,
-    ) -> hal::memory::Barrier<'a, B> {
+        tex: &'a resource::Texture<A>,
+    ) -> hal::TextureBarrier<'a, A> {
         log::trace!("\ttexture -> {:?}", self);
-        let &(ref target, _) = tex.raw.as_ref().expect("Texture is destroyed");
-        let aspects = tex.aspects;
-        hal::memory::Barrier::Image {
-            states: conv::map_texture_state(self.usage.start, aspects)
-                ..conv::map_texture_state(self.usage.end, aspects),
-            target,
-            range: hal::image::SubresourceRange {
-                aspects,
-                level_start: self.selector.levels.start,
-                level_count: Some(self.selector.levels.end - self.selector.levels.start),
-                layer_start: self.selector.layers.start,
-                layer_count: Some(self.selector.layers.end - self.selector.layers.start),
+        let texture = tex.inner.as_raw().expect("Texture is destroyed");
+        hal::TextureBarrier {
+            texture,
+            range: wgt::ImageSubresourceRange {
+                aspect: wgt::TextureAspect::All,
+                base_mip_level: self.selector.levels.start,
+                mip_level_count: NonZeroU32::new(
+                    self.selector.levels.end - self.selector.levels.start,
+                ),
+                base_array_layer: self.selector.layers.start,
+                array_layer_count: NonZeroU32::new(
+                    self.selector.layers.end - self.selector.layers.start,
+                ),
             },
-            families: None,
+            usage: self.usage,
         }
     }
 }
@@ -180,8 +168,8 @@ impl From<PendingTransition<TextureState>> for UsageConflict {
     fn from(e: PendingTransition<TextureState>) -> Self {
         Self::Texture {
             id: e.id.0,
-            mip_levels: e.selector.levels.start as u32..e.selector.levels.end as u32,
-            array_layers: e.selector.layers.start as u32..e.selector.layers.end as u32,
+            mip_levels: e.selector.levels.start..e.selector.levels.end,
+            array_layers: e.selector.layers.start..e.selector.layers.end,
             combined_use: e.usage.end,
         }
     }
@@ -247,10 +235,17 @@ impl<S: ResourceState> ResourceTracker<S> {
         let (index, epoch, backend) = id.0.unzip();
         debug_assert_eq!(backend, self.backend);
         match self.map.entry(index) {
+            // This code explicitly ignores requests for IDs that are no longer valid,
+            // i.e. corresponding to removed entries, or entries that got re-filled
+            // with new elements (having different epochs).
+            // This is needed because of the asynchronous nature of the device internals.
+            // As such, by the time a resource is added to the suspected list, it may
+            // already be fully removed from all the trackers (and be a stale ID).
+            // see https://github.com/gfx-rs/wgpu/issues/1996
             Entry::Occupied(e) => {
-                if e.get().ref_count.load() == 1 {
-                    let res = e.remove();
-                    assert_eq!(res.epoch, epoch, "Epoch mismatch for {:?}", id);
+                // see https://github.com/gfx-rs/wgpu/issues/1996
+                if e.get().epoch == epoch && e.get().ref_count.load() == 1 {
+                    e.remove();
                     true
                 } else {
                     false
@@ -275,8 +270,18 @@ impl<S: ResourceState> ResourceTracker<S> {
             .map(move |(&index, resource)| Valid(S::Id::zip(index, resource.epoch, backend)))
     }
 
+    pub fn get_ref_count(&self, id: Valid<S::Id>) -> &RefCount {
+        let (index, _, _) = id.0.unzip();
+        &self.map[&index].ref_count
+    }
+
+    /// Return true if there is nothing here.
+    pub fn is_empty(&self) -> bool {
+        self.map.is_empty()
+    }
+
     /// Clear the tracked contents.
-    fn clear(&mut self) {
+    pub fn clear(&mut self) {
         self.map.clear();
     }
 
@@ -308,6 +313,7 @@ impl<S: ResourceState> ResourceTracker<S> {
     ///
     /// Returns `Some(Usage)` only if this usage is consistent
     /// across the given selector.
+    #[allow(unused)] // TODO: figure out if this needs to be removed
     pub fn query(&self, id: Valid<S::Id>, selector: S::Selector) -> Option<S::Usage> {
         let (index, epoch, backend) = id.0.unzip();
         debug_assert_eq!(backend, self.backend);
@@ -396,21 +402,6 @@ impl<S: ResourceState> ResourceTracker<S> {
         self.temp.drain(..)
     }
 
-    /// Turn the tracking from the "expand" mode into the "replace" one,
-    /// installing the selected usage as the "first".
-    /// This is a special operation only used by the render pass attachments.
-    pub(crate) fn prepend(
-        &mut self,
-        id: Valid<S::Id>,
-        ref_count: &RefCount,
-        selector: S::Selector,
-        usage: S::Usage,
-    ) -> Result<(), PendingTransition<S>> {
-        Self::get_or_insert(self.backend, &mut self.map, id, ref_count)
-            .state
-            .prepend(id, selector, usage)
-    }
-
     /// Merge another tracker into `self` by extending the current states
     /// without any transitions.
     pub(crate) fn merge_extend(&mut self, other: &Self) -> Result<(), PendingTransition<S>> {
@@ -527,15 +518,6 @@ impl<I: Copy + fmt::Debug + TypedId> ResourceState for PhantomData<I> {
         Ok(())
     }
 
-    fn prepend(
-        &mut self,
-        _id: Valid<Self::Id>,
-        _selector: Self::Selector,
-        _usage: Self::Usage,
-    ) -> Result<(), PendingTransition<Self>> {
-        Ok(())
-    }
-
     fn merge(
         &mut self,
         _id: Valid<Self::Id>,
@@ -557,14 +539,14 @@ pub enum UsageConflict {
     )]
     Buffer {
         id: id::BufferId,
-        combined_use: resource::BufferUse,
+        combined_use: hal::BufferUses,
     },
     #[error("Attempted to use texture {id:?} mips {mip_levels:?} layers {array_layers:?} as a combination of {combined_use:?} within a usage scope.")]
     Texture {
         id: id::TextureId,
         mip_levels: ops::Range<u32>,
         array_layers: ops::Range<u32>,
-        combined_use: resource::TextureUse,
+        combined_use: hal::TextureUses,
     },
 }
 
@@ -599,7 +581,7 @@ impl TrackerSet {
     }
 
     /// Clear all the trackers.
-    pub fn clear(&mut self) {
+    pub fn _clear(&mut self) {
         self.buffers.clear();
         self.textures.clear();
         self.views.clear();
@@ -624,29 +606,14 @@ impl TrackerSet {
         self.query_sets.optimize();
     }
 
-    /// Merge all the trackers of another instance by extending
-    /// the usage. Panics on a stateless conflict, returns a conflict otherwise.
-    pub fn merge_extend_all(&mut self, other: &Self) -> Result<(), UsageConflict> {
+    /// Merge only the stateful trackers of another instance by extending
+    /// the usage. Returns a conflict if any.
+    pub fn merge_extend_stateful(&mut self, other: &Self) -> Result<(), UsageConflict> {
         self.buffers.merge_extend(&other.buffers)?;
         self.textures.merge_extend(&other.textures)?;
-        self.merge_extend_stateless(other);
         Ok(())
     }
 
-    /// Merge all the stateless trackers of another instance by extending
-    /// the usage. Panics on a conflict.
-    pub fn merge_extend_stateless(&mut self, other: &Self) {
-        self.views.merge_extend(&other.views).unwrap();
-        self.bind_groups.merge_extend(&other.bind_groups).unwrap();
-        self.samplers.merge_extend(&other.samplers).unwrap();
-        self.compute_pipes
-            .merge_extend(&other.compute_pipes)
-            .unwrap();
-        self.render_pipes.merge_extend(&other.render_pipes).unwrap();
-        self.bundles.merge_extend(&other.bundles).unwrap();
-        self.query_sets.merge_extend(&other.query_sets).unwrap();
-    }
-
     pub fn backend(&self) -> wgt::Backend {
         self.buffers.backend
     }
