@@ -21,6 +21,7 @@ enum ResourceType {
 
 #[derive(Debug)]
 struct Resource {
+    #[allow(unused)]
     name: Option<String>,
     bind: naga::ResourceBinding,
     ty: ResourceType,
@@ -90,6 +91,7 @@ enum Varying {
     BuiltIn(naga::BuiltIn),
 }
 
+#[allow(unused)]
 #[derive(Debug)]
 struct SpecializationConstant {
     id: u32,
@@ -101,6 +103,7 @@ struct EntryPoint {
     inputs: Vec<Varying>,
     outputs: Vec<Varying>,
     resources: Vec<(naga::Handle<Resource>, GlobalUse)>,
+    #[allow(unused)]
     spec_constants: Vec<SpecializationConstant>,
     sampling_pairs: FastHashSet<(naga::Handle<Resource>, naga::Handle<Resource>)>,
 }
@@ -220,7 +223,7 @@ pub enum InputError {
 pub enum StageError {
     #[error("shader module is invalid")]
     InvalidModule,
-    #[error("unable to find entry point '{0:?}'")]
+    #[error("unable to find entry point '{0}'")]
     MissingEntryPoint(String),
     #[error("shader global {0:?} is not available in the layout pipeline layout")]
     Binding(naga::ResourceBinding, #[source] BindingError),
@@ -790,7 +793,7 @@ impl Interface {
         list: &mut Vec<Varying>,
         binding: Option<&naga::Binding>,
         ty: naga::Handle<naga::Type>,
-        arena: &naga::Arena<naga::Type>,
+        arena: &naga::UniqueArena<naga::Type>,
     ) {
         let numeric_ty = match arena[ty].inner {
             naga::TypeInner::Scalar { kind, width } => NumericType {
@@ -819,7 +822,11 @@ impl Interface {
                 return;
             }
             ref other => {
-                log::error!("Unexpected varying type: {:?}", other);
+                //Note: technically this should be at least `log::error`, but
+                // the reality is - every shader coming from `glslc` outputs an array
+                // of clip distances and hits this path :(
+                // So we lower it to `log::warn` to be less annoying.
+                log::warn!("Unexpected varying type: {:?}", other);
                 return;
             }
         };
@@ -923,7 +930,7 @@ impl Interface {
             entry_points.insert((entry_point.stage, entry_point.name.clone()), ep);
         }
 
-        Interface {
+        Self {
             features,
             resources,
             entry_points,
